from py42.services._auth import C42RenewableAuth


class StorageTmpAuth(C42RenewableAuth):
    def __init__(self):
        super(StorageTmpAuth, self).__init__()
        self._storage_url = None

    def get_storage_url(self):
        self.get_credentials()
        return self._server_url

    def get_tmp_auth(self):
        raise NotImplementedError()

    def _get_credentials(self):
        login_info = self.get_tmp_auth()
        login_token = login_info[u"loginToken"]
        self._server_url = login_info[u"serverUrl"]
        return u"login_token {}".format(login_token)


class FileArchiveTmpAuth(StorageTmpAuth):
    def __init__(self, connection, user_id, device_guid, destination_guid):
        super(FileArchiveTmpAuth, self).__init__()
        self._connection = connection
        self._user_id = user_id
        self._device_guid = device_guid
        self._destination_guid = destination_guid

    def get_tmp_auth(self):
        uri = u"/api/LoginToken"
        data = {
            u"userId": self._user_id,
            u"sourceGuid": self._device_guid,
            u"destinationGuid": self._destination_guid,
        }
        response = self._connection.post(uri, json=data)
        return response


class SecurityArchiveTmpAuth(StorageTmpAuth):
    def __init__(self, connection, plan_uid, destination_guid):
        super(SecurityArchiveTmpAuth, self).__init__()
        self._connection = connection
        self._plan_uid = plan_uid
        self._destination_guid = destination_guid

    def get_tmp_auth(self):
        uri = u"/api/StorageAuthToken"
        data = {u"planUid": self._plan_uid, u"destinationGuid": self._destination_guid}
        response = self._connection.post(uri, json=data)
        return response


class V1Auth(C42RenewableAuth):
    def __init__(self, storage_tmp_session):
        super(V1Auth, self).__init__()
        self._auth_session = storage_tmp_session

    def _get_credentials(self):
        uri = u"/api/AuthToken"
        response = self._auth_session.post(uri)
        return u"token {}-{}".format(response[0], response[1])



from py42.exceptions import Py42BadRequestError
from py42.exceptions import Py42NotFoundError
from py42.services import BaseService


class DetectionListUserService(BaseService):
    """Administrator utility to manage High Risk employees information.

    `Support Documentation <https://support.code42.com/Administrator/Cloud/Monitoring_and_managing/Detection_list_management_APIs>`__
    """

    _api_version = u"v2"
    _uri_prefix = u"/svc/api/{}".format(_api_version)
    _resource = u"/user"

    def __init__(self, connection, user_context, user_service):
        super(DetectionListUserService, self).__init__(connection)
        self._user_context = user_context
        self._user_service = user_service

    def _make_uri(self, action):
        return u"{}{}{}".format(self._uri_prefix, self._resource, action)

    def create_if_not_exists(self, user_id):
        """Find out whether the detection list profile exists for a given uid. If not,
           if user_id is a valid uid of code42 user then it will create detection list profile
           of the user.

            Returns True when profile is created else raises error.

            Args:
                user_id (str or int): Uid of user.

            Returns:
                bool
        """
        try:
            self.get_by_id(user_id)
        except (Py42NotFoundError, Py42BadRequestError):
            user = self._user_service.get_by_uid(user_id)
            self.create(user[u"username"])
        return True

    def create(self, username):
        """Create a detection list profile for a user.

        Args:
            username (str): Username of the user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"userName": username,
            u"notes": "",
            u"riskFactors": [],
            u"cloudUsernames": [],
        }
        uri = self._make_uri(u"/create")
        return self._connection.post(uri, json=data)

    def get_by_id(self, user_id):
        """Get user details by user id.

        Args:
            user_id (str or int): Id of the user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"userId": user_id,
        }
        uri = self._make_uri(u"/getbyid")
        return self._connection.post(uri, json=data)

    def get(self, username):
        """Get user details by username.

        Args:
            username (str): Username of the user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"username": username,
        }
        uri = self._make_uri(u"/getbyusername")
        return self._connection.post(uri, json=data)

    def update_notes(self, user_id, notes):
        """Add or update notes related to the user.

        Args:
            user_id (str or int): The user_id whose notes need to be updated.
            notes (str): User profile notes.

        Returns:
            :class:`py42.response.Py42Response`
        """
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"userId": user_id,
            u"notes": notes,
        }
        uri = self._make_uri(u"/updatenotes")
        return self._connection.post(uri, json=data)

    def add_risk_tags(self, user_id, tags):
        """Add one or more tags.

        Args:
            user_id (str or int): The user_id whose tag(s) needs to be updated.
            tags (str or list of str ): A single tag or multiple tags in a list to be added.
               e.g "tag1" or ["tag1", "tag2"]

        Returns:
            :class:`py42.response.Py42Response`
        """

        if not isinstance(tags, (list, tuple)):
            tags = [tags]

        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"userId": user_id,
            u"riskFactors": tags,
        }
        uri = self._make_uri(u"/addriskfactors")
        return self._connection.post(uri, json=data)

    def remove_risk_tags(self, user_id, tags):
        """Remove one or more tags.Args:

        Args:
            user_id (str or int): The user_id whose tag(s) needs to be removed.
            tags (str or list of str ): A single tag or multiple tags in a list to be removed.
               e.g "tag1" or ["tag1", "tag2"].

        Returns:
            :class:`py42.response.Py42Response`
        """
        if not isinstance(tags, (list, tuple)):
            tags = [tags]

        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"userId": user_id,
            u"riskFactors": tags,
        }
        uri = self._make_uri(u"/removeriskfactors")
        return self._connection.post(uri, json=data)

    def add_cloud_alias(self, user_id, alias):
        """Add a cloud alias.

        Args:
            user_id (str or int): The user_id whose alias needs to be updated.
            alias (str): An alias to be added.

        Returns:
            :class:`py42.response.Py42Response`
        """
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"userId": user_id,
            u"cloudUsernames": [alias],
        }
        uri = self._make_uri(u"/addcloudusernames")
        return self._connection.post(uri, json=data)

    def remove_cloud_alias(self, user_id, alias):
        """Remove one or more cloud alias.

        Args:
            user_id (str or int): The user_id whose alias needs to be removed.
            alias (str): An alias to be removed.

        Returns:
            :class:`py42.response.Py42Response`
        """
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"userId": user_id,
            u"cloudUsernames": [alias],
        }
        uri = self._make_uri(u"/removecloudusernames")
        return self._connection.post(uri, json=data)

    def refresh(self, user_id):
        """Refresh SCIM attributes of a user.

        Args:
            user_id (str or int): The user_id of the user whose attributes need to be refreshed.

        Returns:
            :class:`py42.response.Py42Response`
        """
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"userId": user_id,
        }
        uri = self._make_uri(u"/refresh")
        return self._connection.post(uri, json=data)



import py42.settings as settings


def get_all_pages(func, key, *args, **kwargs):
    item_count = settings.items_per_page
    page_num = 0
    if kwargs.get("page_size") is None:
        kwargs[u"page_size"] = settings.items_per_page
    while item_count >= settings.items_per_page:
        page_num += 1
        response = func(*args, page_num=page_num, **kwargs)
        yield response
        page_items = response[key]
        item_count = len(page_items)



from requests.auth import HTTPBasicAuth

from py42.clients import Clients
from py42.clients._archive_access import ArchiveAccessorManager
from py42.clients.alertrules import AlertRulesClient
from py42.clients.alerts import AlertsClient
from py42.clients.archive import ArchiveClient
from py42.clients.authority import AuthorityClient
from py42.clients.detectionlists import DetectionListsClient
from py42.clients.securitydata import SecurityDataClient
from py42.services import Services
from py42.services._auth import V3Auth
from py42.services._connection import Connection
from py42.services._keyvaluestore import KeyValueStoreClient
from py42.services.administration import AdministrationService
from py42.services.alertrules import AlertRulesService
from py42.services.alerts import AlertService
from py42.services.archive import ArchiveService
from py42.services.detectionlists.departing_employee import DepartingEmployeeService
from py42.services.detectionlists.high_risk_employee import HighRiskEmployeeService
from py42.services.detectionlists.user_profile import DetectionListUserService
from py42.services.devices import DeviceService
from py42.services.fileevent import FileEventService
from py42.services.legalhold import LegalHoldService
from py42.services.orgs import OrgService
from py42.services.preservationdata import PreservationDataService
from py42.services.savedsearch import SavedSearchService
from py42.services.securitydata import SecurityDataService
from py42.services.storage._service_factory import ConnectionManager
from py42.services.storage._service_factory import StorageServiceFactory
from py42.services.users import UserService
from py42.usercontext import UserContext


def from_local_account(host_address, username, password):
    """Creates a :class:`~py42.sdk.SDKClient` object for accessing the Code42 REST APIs using the
    supplied credentials. Currently, only accounts created within the Code42 console or using the
    APIs (including py42) are supported. Username/passwords that are based on Active Directory,
    Okta, or other Identity providers cannot be used with this method.

    Args:
        host_address (str): The domain name of the Code42 instance being authenticated to, e.g.
            console.us.code42.com
        username (str): The username of the authenticating account.
        password (str): The password of the authenticating account.

    Returns:
        :class:`py42.sdk.SDKClient`
    """
    client = SDKClient.from_local_account(host_address, username, password)

    # test credentials
    client.users.get_current()
    return client


class SDKClient(object):
    def __init__(self, main_connection, auth):
        services, user_ctx = _init_services(main_connection, auth)
        self._clients = _init_clients(services, main_connection)
        self._user_ctx = user_ctx

    @classmethod
    def from_local_account(cls, host_address, username, password):
        """Creates a :class:`~py42.sdk.SDKClient` object for accessing the Code42 REST APIs using
        the supplied credentials. Currently, only accounts created within the Code42 console or
        using the APIs (including py42) are supported. Username/passwords that are based on Active
        Directory, Okta, or other Identity providers cannot be used with this method.

        Args:
            host_address (str): The domain name of the Code42 instance being authenticated to, e.g.
                console.us.code42.com
            username (str): The username of the authenticating account.
            password (str): The password of the authenticating account.

        Returns:
            :class:`py42.sdk.SDKClient`
        """
        basic_auth = None
        if username and password:
            basic_auth = HTTPBasicAuth(username, password)
        auth_connection = Connection.from_host_address(host_address, auth=basic_auth)
        v3_auth = V3Auth(auth_connection)
        main_connection = Connection.from_host_address(host_address, auth=v3_auth)

        return cls(main_connection, v3_auth)

    @property
    def serveradmin(self):
        """A collection of methods for getting server information for on-premise environments
        and tenant information for cloud environments.

        Returns:
            :class:`py42.services.administration.AdministrationService`
        """
        return self._clients.authority.administration

    @property
    def archive(self):
        """A collection of methods for accessing Code42 storage archives. Useful for doing
        web-restores or finding a file on an archive.

        Returns:
            :class:`py42.services.archive.ArchiveClient`
        """
        return self._clients.archive

    @property
    def users(self):
        """A collection of methods for retrieving or updating data about users in the Code42
        environment.

        Returns:
            :class:`py42.services.users.UserService`
        """
        return self._clients.authority.users

    @property
    def devices(self):
        """A collection of methods for retrieving or updating data about devices in the Code42
        environment.

        Returns:
            :class:`py42.services.devices.DeviceService`
        """
        return self._clients.authority.devices

    @property
    def orgs(self):
        """A collection of methods for retrieving or updating data about organizations in the
        Code42 environment.

        Returns:
            :class:`py42.services.orgs.OrgService`
        """
        return self._clients.authority.orgs

    @property
    def legalhold(self):
        """A collection of methods for retrieving and updating legal-hold matters, policies, and
        custodians.

        Returns:
            :class:`py42.services.legalhold.LegalHoldService`
        """
        return self._clients.authority.legalhold

    @property
    def usercontext(self):
        """A collection of methods related to getting information about the currently logged in
        user, such as the tenant ID.

        Returns:
            :class:`py42.usercontext.UserContext`
        """
        return self._user_ctx

    @property
    def securitydata(self):
        """A collection of methods and properties for getting security data such as:
            * File events
            * Alerts
            * Security plan information

        Returns:
            :class:`py42.services.securitydata.SecurityDataClient`
        """
        return self._clients.securitydata

    @property
    def detectionlists(self):
        """A collection of properties each containing methods for managing specific detection
        lists, such as departing employees.

        Returns:
            :class:`py42.services.detectionlists.DetectionListsClient`
        """
        return self._clients.detectionlists

    @property
    def alerts(self):
        """A collection of methods related to retrieving and updating alerts rules.

        Returns:
            :class:`py42.services.alertrules.AlertRulesClient`
        """
        return self._clients.alerts


def _init_services(main_connection, main_auth):
    alert_rules_key = u"FedObserver-API_URL"
    alerts_key = u"AlertService-API_URL"
    file_events_key = u"FORENSIC_SEARCH-API_URL"
    preservation_data_key = u"PRESERVATION-DATA-SERVICE_API-URL"
    employee_case_mgmt_key = u"employeecasemanagement-API_URL"
    kv_prefix = u"simple-key-value-store"

    kv_connection = Connection.from_microservice_prefix(main_connection, kv_prefix)
    kv_service = KeyValueStoreClient(kv_connection)

    alert_rules_conn = Connection.from_microservice_key(
        kv_service, alert_rules_key, auth=main_auth
    )
    alerts_conn = Connection.from_microservice_key(
        kv_service, alerts_key, auth=main_auth
    )
    file_events_conn = Connection.from_microservice_key(
        kv_service, file_events_key, auth=main_auth
    )
    pds_conn = Connection.from_microservice_key(
        kv_service, preservation_data_key, auth=main_auth
    )
    ecm_conn = Connection.from_microservice_key(
        kv_service, employee_case_mgmt_key, auth=main_auth
    )

    user_svc = UserService(main_connection)
    administration_svc = AdministrationService(main_connection)
    file_event_svc = FileEventService(file_events_conn)
    user_ctx = UserContext(administration_svc)
    user_profile_svc = DetectionListUserService(ecm_conn, user_ctx, user_svc)

    services = Services(
        administration=administration_svc,
        archive=ArchiveService(main_connection),
        devices=DeviceService(main_connection),
        legalhold=LegalHoldService(main_connection),
        orgs=OrgService(main_connection),
        securitydata=SecurityDataService(main_connection),
        users=UserService(main_connection),
        alertrules=AlertRulesService(alert_rules_conn, user_ctx, user_profile_svc),
        alerts=AlertService(alerts_conn, user_ctx),
        fileevents=file_event_svc,
        savedsearch=SavedSearchService(file_events_conn, file_event_svc),
        preservationdata=PreservationDataService(pds_conn),
        departingemployee=DepartingEmployeeService(
            ecm_conn, user_ctx, user_profile_svc
        ),
        highriskemployee=HighRiskEmployeeService(ecm_conn, user_ctx, user_profile_svc),
        userprofile=user_profile_svc,
    )

    return services, user_ctx


def _init_clients(services, connection):
    authority = AuthorityClient(
        administration=services.administration,
        archive=services.archive,
        devices=services.devices,
        legalhold=services.legalhold,
        orgs=services.orgs,
        securitydata=services.securitydata,
        users=services.users,
    )
    detectionlists = DetectionListsClient(
        services.userprofile, services.departingemployee, services.highriskemployee
    )

    storage_service_factory = StorageServiceFactory(
        connection, services.devices, ConnectionManager()
    )
    alertrules = AlertRulesClient(services.alerts, services.alertrules)
    securitydata = SecurityDataClient(
        services.securitydata,
        services.fileevents,
        services.preservationdata,
        services.savedsearch,
        storage_service_factory,
    )
    alerts = AlertsClient(services.alerts, alertrules)
    archive_accessor_mgr = ArchiveAccessorManager(
        services.archive, storage_service_factory
    )
    archive = ArchiveClient(archive_accessor_mgr, services.archive)
    clients = Clients(
        authority=authority,
        detectionlists=detectionlists,
        alerts=alerts,
        securitydata=securitydata,
        archive=archive,
    )
    return clients



import posixpath
import time
from collections import namedtuple

from py42.exceptions import Py42ArchiveFileNotFoundError
from py42.settings import debug
from py42.util import format_dict


FileSelection = namedtuple(u"FileSelection", u"path_set, num_files, num_dirs, size")


class FileType(object):
    DIRECTORY = u"directory"
    FILE = u"file"


class ArchiveAccessorManager(object):
    def __init__(self, archive_service, storage_service_factory):
        self._archive_service = archive_service
        self._storage_service_factory = storage_service_factory

    def get_archive_accessor(
        self,
        device_guid,
        destination_guid=None,
        private_password=None,
        encryption_key=None,
    ):
        service = self._storage_service_factory.create_archive_service(
            device_guid, destination_guid=destination_guid
        )
        decryption_keys = self._get_decryption_keys(
            device_guid, private_password, encryption_key
        )
        session_id = self._create_restore_session(
            service, device_guid, **decryption_keys
        )
        restore_job_manager = create_restore_job_manager(
            service, device_guid, session_id
        )
        file_size_poller = create_file_size_poller(service, device_guid)
        return ArchiveAccessor(
            device_guid, session_id, service, restore_job_manager, file_size_poller,
        )

    def _get_decryption_keys(self, device_guid, private_password, encryption_key):
        decryption_keys = {}
        if encryption_key:
            decryption_keys[u"encryption_key"] = encryption_key
        else:
            data_key_token = (
                self._get_data_key_token(device_guid) if not encryption_key else None
            )
            if data_key_token:
                decryption_keys[u"data_key_token"] = data_key_token

            if private_password:
                decryption_keys[u"private_password"] = private_password
        return decryption_keys

    def _get_data_key_token(self, device_guid):
        return self._archive_service.get_data_key_token(device_guid)[u"dataKeyToken"]

    @staticmethod
    def _create_restore_session(storage_archive_service, device_guid, **kwargs):
        response = storage_archive_service.create_restore_session(device_guid, **kwargs)
        return response[u"webRestoreSessionId"]


def _create_file_selections(file_paths, metadata_list, file_sizes=None):
    file_selections = []
    for i in range(0, len(file_paths)):
        metadata = metadata_list[i]
        size_info = file_sizes[i] if file_sizes else _get_default_file_size()
        path_set = {
            u"type": metadata[u"type"],
            u"path": metadata[u"path"],
            u"selected": True,
        }
        selection = FileSelection(
            path_set, size_info[u"numFiles"], size_info[u"numDirs"], size_info[u"size"],
        )
        file_selections.append(selection)

    return file_selections


class ArchiveAccessor(object):

    DEFAULT_DIRECTORY_DOWNLOAD_NAME = u"download"
    JOB_POLLING_INTERVAL = 1

    def __init__(
        self,
        device_guid,
        archive_session_id,
        storage_archive_service,
        restore_job_manager,
        file_size_poller,
    ):
        self._device_guid = device_guid
        self._archive_session_id = archive_session_id
        self._storage_archive_service = storage_archive_service
        self._restore_job_manager = restore_job_manager
        self._file_size_poller = file_size_poller

    def stream_from_backup(self, file_paths, file_size_calc_timeout=None):
        file_selections = self._create_file_selections(
            file_paths, file_size_calc_timeout
        )
        return self._restore_job_manager.get_stream(file_selections)

    def _create_file_selections(self, file_paths, file_size_calc_timeout):
        if not isinstance(file_paths, (list, tuple)):
            file_paths = [file_paths]
        file_paths = [fp.replace(u"\\", u"/") for fp in file_paths]
        metadata_list = self._get_restore_metadata(file_paths)
        file_ids = [md[u"id"] for md in metadata_list]
        file_sizes = self._file_size_poller.get_file_sizes(
            file_ids, timeout=file_size_calc_timeout
        )
        return _create_file_selections(file_paths, metadata_list, file_sizes)

    def _get_restore_metadata(self, file_paths):
        metadata_list = []
        for path in file_paths:
            metadata = self._get_file_via_walking_tree(path)
            metadata_list_entry = {
                u"id": metadata[u"id"],
                u"path": metadata[u"path"],
                u"type": metadata[u"type"],
            }
            metadata_list.append(metadata_list_entry)
        return metadata_list

    def _get_file_via_walking_tree(self, file_path):
        path_parts = file_path.split(u"/")
        path_root = path_parts[0] + u"/"

        response = self._get_children(node_id=None)
        for root in response:
            if root[u"path"].lower() == path_root.lower():
                return self._walk_tree(response, root, path_parts[1:])

        raise Py42ArchiveFileNotFoundError(response, self._device_guid, file_path)

    def _walk_tree(self, response, current_node, remaining_path_components):
        if not remaining_path_components or not remaining_path_components[0]:
            return current_node

        children = self._get_children(node_id=current_node[u"id"])
        current_node_path = current_node[u"path"]
        target_child_path = posixpath.join(
            current_node_path, remaining_path_components[0]
        )

        for child in children:
            if child[u"path"].lower() == target_child_path.lower():
                return self._walk_tree(response, child, remaining_path_components[1:])

        raise Py42ArchiveFileNotFoundError(
            response, self._device_guid, target_child_path
        )

    def _get_children(self, node_id=None):
        return self._storage_archive_service.get_file_path_metadata(
            self._archive_session_id,
            self._device_guid,
            file_id=node_id,
            show_deleted=True,
        )


def _get_default_file_size():
    return {u"numFiles": 1, u"numDirs": 1, u"size": 1}


class _RestorePoller(object):
    JOB_POLLING_INTERVAL_SECONDS = 1

    def __init__(self, storage_archive_service, device_guid, job_polling_interval=None):
        self._storage_archive_service = storage_archive_service
        self._device_guid = device_guid
        self._job_polling_interval = (
            job_polling_interval or self.JOB_POLLING_INTERVAL_SECONDS
        )


def _create_size_dict(job_id, size_response):
    size_dict = size_response.data
    size_dict[u"jobId"] = job_id
    return size_dict


def _print_file_size(size_dict):
    debug.logger.debug(format_dict(size_dict))


class FileSizePoller(_RestorePoller):
    def __init__(
        self, storage_archive_service, device_guid, job_polling_interval=None,
    ):
        super(FileSizePoller, self).__init__(
            storage_archive_service, device_guid, job_polling_interval
        )

    def get_file_sizes(self, file_ids, timeout):
        if not timeout:
            # Skips file size calculation
            return None

        job_ids = self._start_poll(file_ids)
        return self._wait_for_jobs(job_ids, timeout)

    def _start_poll(self, file_ids):
        job_ids = []
        for file_id in file_ids:
            response = self._storage_archive_service.create_file_size_job(
                self._device_guid, file_id
            )
            job_id = response[u"jobId"]
            job_ids.append(job_id)
        return job_ids

    def _wait_for_jobs(self, job_ids, timeout):
        t0 = time.time()
        sizes = []

        # Waits until the job_ids stack is empty
        while job_ids:
            for job_id in job_ids:
                response = self._get_job_status(job_id)
                size_dict = _create_size_dict(job_id, response)
                _print_file_size(size_dict)
                if response[u"status"].lower() == u"done":
                    job_ids.remove(job_id)
                    sizes.append(size_dict)

            # File size calculation is taking too long.
            if time.time() - t0 > timeout:
                return None

            if job_ids:
                time.sleep(self._job_polling_interval)
        return sizes

    def _get_job_status(self, job_id):
        return self._storage_archive_service.get_file_size_job(
            job_id, self._device_guid
        )


class RestoreJobManager(_RestorePoller):
    def __init__(
        self,
        storage_archive_service,
        device_guid,
        archive_session_id,
        job_polling_interval=None,
    ):
        super(RestoreJobManager, self).__init__(
            storage_archive_service, device_guid, job_polling_interval
        )
        self._archive_session_id = archive_session_id

    def get_stream(self, file_selections):
        response = self._start_restore(file_selections)
        job_id = response["jobId"]
        self._wait_for_job(job_id)
        return self._get_stream(job_id)

    def _wait_for_job(self, job_id):
        while not self._is_job_complete(job_id):
            time.sleep(self._job_polling_interval)

    def _is_job_complete(self, job_id):
        response = self._storage_archive_service.get_restore_status(job_id)
        is_done = response[u"done"]
        percentage_dict = {
            u"jobId": job_id,
            u"status": response.data.get("status"),
            u"percentComplete": response[u"percentComplete"] if not is_done else 100,
        }
        debug.logger.debug(format_dict(percentage_dict))
        return is_done

    def _start_restore(self, file_selections):
        num_files = sum([fs.num_files for fs in file_selections])
        num_dirs = sum([fs.num_dirs for fs in file_selections])
        size = sum([fs.size for fs in file_selections])
        zip_result = _check_for_multiple_files(file_selections) or None
        return self._storage_archive_service.start_restore(
            guid=self._device_guid,
            web_restore_session_id=self._archive_session_id,
            path_set=[fs.path_set for fs in file_selections],
            num_files=num_files,
            num_dirs=num_dirs,
            size=size,
            zip_result=zip_result,
            show_deleted=True,
        )

    def _get_stream(self, job_id):
        response = self._storage_archive_service.stream_restore_result(job_id)
        return response


def create_restore_job_manager(
    storage_archive_service, device_guid, archive_session_id
):
    return RestoreJobManager(storage_archive_service, device_guid, archive_session_id)


def create_file_size_poller(storage_archive_service, device_guid):
    return FileSizePoller(storage_archive_service, device_guid)


def _check_for_multiple_files(file_selection):
    if not file_selection:
        return False
    if len(file_selection) > 1:
        return True

    # Only one file selected
    selection = file_selection[0]
    return selection.path_set[u"type"].lower() == u"directory"



from threading import Lock

from py42._compat import str
from py42.exceptions import Py42StorageSessionInitializationError
from py42.services._connection import Connection
from py42.services.storage._auth import FileArchiveTmpAuth
from py42.services.storage._auth import SecurityArchiveTmpAuth
from py42.services.storage.archive import StorageArchiveService
from py42.services.storage.preservationdata import StoragePreservationDataService
from py42.services.storage.securitydata import StorageSecurityDataService


class StorageServiceFactory(object):
    def __init__(self, connection, device_service, connection_manager):
        self._connection = connection
        self._device_service = device_service
        self._connection_manager = connection_manager

    def create_archive_service(self, device_guid, destination_guid=None):
        if destination_guid is None:
            destination_guid = self._auto_select_destination_guid(device_guid)

        auth = FileArchiveTmpAuth(
            self._connection, u"my", device_guid, destination_guid
        )
        connection = self._connection_manager.get_storage_connection(auth)
        return StorageArchiveService(connection)

    def create_security_data_service(self, plan_uid, destination_guid):
        auth = SecurityArchiveTmpAuth(self._connection, plan_uid, destination_guid)
        connection = self._connection_manager.get_storage_connection(auth)
        return StorageSecurityDataService(connection)

    def create_preservation_data_service(self, host_address):
        main_connection = self._connection.clone(host_address)
        streaming_connection = Connection.from_host_address(host_address)
        return StoragePreservationDataService(main_connection, streaming_connection)

    def _auto_select_destination_guid(self, device_guid):
        response = self._device_service.get_by_guid(
            device_guid, include_backup_usage=True
        )
        # take the first destination guid we find
        destination_list = response["backupUsage"]
        if not destination_list:
            raise Exception(
                u"No destinations found for device guid: {}".format(device_guid)
            )
        return destination_list[0][u"targetComputerGuid"]


class ConnectionManager(object):
    def __init__(self, session_cache=None):
        self._session_cache = session_cache or {}
        self._list_update_lock = Lock()

    def get_saved_connection_for_url(self, url):
        return self._session_cache.get(url.lower())

    def get_storage_connection(self, tmp_auth):
        try:
            url = tmp_auth.get_storage_url()
            connection = self.get_saved_connection_for_url(url)
            if connection is None:
                with self._list_update_lock:
                    connection = self.get_saved_connection_for_url(url)
                    if connection is None:
                        connection = Connection.from_host_address(url, auth=tmp_auth)
                        self._session_cache[url.lower()] = connection
        except Exception as ex:
            message = u"Failed to create or retrieve connection, caused by: {}".format(
                str(ex)
            )
            raise Py42StorageSessionInitializationError(ex, message)
        return connection



from py42 import settings
from py42.services import BaseService
from py42.services.util import get_all_pages


class DeviceService(BaseService):
    """A class to interact with Code42 device/computer APIs."""

    def get_page(
        self,
        page_num,
        active=None,
        blocked=None,
        org_uid=None,
        user_uid=None,
        destination_guid=None,
        include_backup_usage=None,
        include_counts=True,
        page_size=None,
        q=None,
    ):
        """Gets a page of devices.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Computer-get>`__

        Args:
            page_num (int): The page number to request.
            active (bool, optional): Filters results by device state. When set to True, gets all
                active devices. When set to False, gets all deactivated devices. When set to None
                or excluded, gets all devices regardless of state. Defaults to None.
            blocked (bool, optional): Filters results by blocked status: True or False. Defaults
                to None.
            org_uid (int, optional): The identification number of an Organization. Defaults to None.
            user_uid (int, optional): The identification number of a User. Defaults to None.
            destination_guid (str or int, optional): The globally unique identifier of the storage
                server that the device back up to. Defaults to None.
            include_backup_usage (bool, optional): A flag to denote whether to include the
                destination and its backup stats. Defaults to None.
            include_counts (bool, optional): A flag to denote whether to include total, warning,
                and critical counts. Defaults to True.
            page_size (int, optional): The number of devices to return per page. Defaults to
                `py42.settings.items_per_page`.
            q (str, optional): Searches results flexibly by incomplete GUID, hostname,
                computer name, etc. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """

        uri = u"/api/Computer"
        page_size = page_size or settings.items_per_page
        params = {
            u"active": active,
            u"blocked": blocked,
            u"orgUid": org_uid,
            u"userUid": user_uid,
            u"targetComputerGuid": destination_guid,
            u"incBackupUsage": include_backup_usage,
            u"incCounts": include_counts,
            u"pgNum": page_num,
            u"pgSize": page_size,
            u"q": q,
        }

        return self._connection.get(uri, params=params)

    def get_all(
        self,
        active=None,
        blocked=None,
        org_uid=None,
        user_uid=None,
        destination_guid=None,
        include_backup_usage=None,
        include_counts=True,
        q=None,
        **kwargs
    ):
        """Gets all device information.

        When no arguments are passed, all records are returned. To filter results, specify
        respective arguments. For example, to retrieve all active and blocked devices, pass
        active=true and blocked=true.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Computer-get>`__

        Args:
            active (bool, optional): Filters results by device state. When set to True, gets all
                active devices. When set to False, gets all deactivated devices. When set to None
                or excluded, gets all devices regardless of state. Defaults to None.
            blocked (bool, optional): Filters results by blocked status: True or False. Defaults
                to None.
            org_uid (int, optional): The identification number of an Organization. Defaults to None.
            user_uid (int, optional): The identification number of a User. Defaults to None.
            destination_guid (str or int, optional): The globally unique identifier of the storage
                server that the device back up to. Defaults to None.
            include_backup_usage (bool, optional): A flag to denote whether to include the
                destination and its backup stats. Defaults to None.
            include_counts (bool, optional): A flag to denote whether to include total, warning,
                and critical counts. Defaults to True.
            q (str, optional): Searches results flexibly by incomplete GUID, hostname,
                computer name, etc. Defaults to None.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of devices.

            The devices returned by `get_all()` are based on the role and permissions of the user
            authenticating the py42 SDK.
        """

        return get_all_pages(
            self.get_page,
            u"computers",
            active=active,
            blocked=blocked,
            org_uid=org_uid,
            user_uid=user_uid,
            destination_guid=destination_guid,
            include_backup_usage=include_backup_usage,
            include_counts=include_counts,
            q=q,
            **kwargs
        )

    def get_by_id(self, device_id, include_backup_usage=None, **kwargs):
        """Gets device information by ID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Computer-get>`__

        Args:
            device_id (int): The identification number of the device.
            include_backup_usage (bool, optional): A flag to denote whether to include the
                destination and its backup stats. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`: A response containing device information.
        """
        uri = u"/api/Computer/{}".format(device_id)
        params = dict(incBackupUsage=include_backup_usage, **kwargs)
        return self._connection.get(uri, params=params)

    def get_by_guid(self, guid, include_backup_usage=None, **kwargs):
        """Gets device information by GUID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Computer-get>`__

        Args:
            guid (str): The globally unique identifier of the device.
            include_backup_usage (bool, optional): A flag to denote whether to include the
                destination and its backup stats. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`: A response containing device information.
        """
        uri = u"/api/Computer/{}".format(guid)
        params = dict(idType=u"guid", incBackupUsage=include_backup_usage, **kwargs)
        return self._connection.get(uri, params=params)

    def block(self, device_id):
        """Blocks a device causing the user not to be able to log in to or restore from Code42 on
        that device.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#ComputerBlock>`__

        Args:
            device_id (int): The identification number of the device.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/ComputerBlock/{}".format(device_id)
        return self._connection.put(uri)

    def unblock(self, device_id):
        """Unblocks a device, permitting a user to be able to login and restore again.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#ComputerBlock>`__

        Args:
            device_id (int): The identification number of the device.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/ComputerBlock/{}".format(device_id)
        return self._connection.delete(uri)

    def deactivate(self, device_id):
        """Deactivates a device, causing backups to stop and archives to go to cold storage.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#ComputerDeactivation>`__

        Args:
            device_id (int): The identification number of the device.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/v4/computer-deactivation/update"
        data = {u"id": device_id}
        return self._connection.post(uri, json=data)

    def reactivate(self, device_id):
        """Activates a previously deactivated device.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#ComputerDeactivation>`__

        Args:
            device_id (int): The identification number of the device.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/v4/computer-deactivation/remove"
        data = {u"id": device_id}
        return self._connection.post(uri, json=data)

    def deauthorize(self, device_id):
        """Deauthorizes the device with the given ID. If used on a cloud connector device, it will
        remove the authorization token for that account.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#ComputerDeauthorization>`__

        Args:
            device_id (int): The identification number of the device.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/ComputerDeauthorization/{}".format(device_id)
        return self._connection.put(uri)

    def get_settings(self, guid, keys=None):
        """Gets settings of the device.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#DeviceSetting>`__

        Args:
            guid (str): The globally unique identifier of the device.
            keys (str, optional): A comma separated list of device keys. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`: A response containing settings information.
        """
        uri = u"/api/v4/device-setting/view"
        params = {u"guid": guid, u"keys": keys}
        return self._connection.get(uri, params=params)

    def get_agent_state(self, guid, property_name):
        """Gets the agent state of the device.
            `REST Documentation <https://console.us.code42.com/swagger/index.html?urls.primaryName=v14#/agent-state/AgentState_ViewByDeviceGuid>`__

            Args:
                guid (str): The globally unique identifier of the device.
                property_name (str): The name of the property to retrieve (e.g. `fullDiskAccess`).

            Returns:
                :class:`py42.response.Py42Response`: A response containing settings information.
            """
        uri = u"/api/v14/agent-state/view-by-device-guid"
        params = {u"deviceGuid": guid, u"propertyName": property_name}
        return self._connection.get(uri, params=params)

    def get_agent_full_disk_access_state(self, guid):
        """Gets the full disk access status of a device.
            `REST Documentation <https://console.us.code42.com/swagger/index.html?urls.primaryName=v14#/agent-state/AgentState_ViewByDeviceGuid>`__

            Args:
                guid (str): The globally unique identifier of the device.

            Returns:
                :class:`py42.response.Py42Response`: A response containing settings information.
            """
        return self.get_agent_state(guid, u"fullDiskAccess")



from py42._compat import str
from py42.sdk.queries import BaseQuery


class AlertQuery(BaseQuery):
    """Helper class for building Code42 Alert queries.

    An AlertQuery instance's ``all()`` and ``any()`` take one or more
    :class:`~py42.sdk.queries.query_filter.FilterGroup` objects to construct a query that
    can be passed to the :meth:`AlertService.search()` method. ``all()`` returns results
    that match all of the provided filter criteria, ``any()`` will return results that
    match any of the filters.

    For convenience, the :class:`AlertQuery` constructor does the same as ``all()``.

    Usage example::

        state_filter = AlertState.eq(AlertState.OPEN)
        rule_name_filter = RuleName.contains("EmailRule")
        query = AlertQuery.all(state_filter, rule_name_filter)
    """

    def __init__(self, *args, **kwargs):
        super(AlertQuery, self).__init__(*args, **kwargs)
        self.sort_key = u"CreatedAt"
        self.page_number = 0
        self.page_size = 500
        self.sort_direction = u"desc"

    def __str__(self):
        groups_string = u",".join(
            str(group_item) for group_item in self._filter_group_list
        )
        json = u'{{"tenantId": null, "groupClause":"{0}", "groups":[{1}], "pgNum":{2}, "pgSize":{3}, "srtDirection":"{4}", "srtKey":"{5}"}}'.format(
            self._group_clause,
            groups_string,
            self.page_number,
            self.page_size,
            self.sort_direction,
            self.sort_key,
        )
        return json

    def __iter__(self):
        filter_group_list = [dict(item) for item in self._filter_group_list]
        output_dict = {
            u"tenantId": None,
            u"groupClause": self._group_clause,
            u"groups": filter_group_list,
            u"pgNum": self.page_number,
            u"pgSize": self.page_size,
            u"srtDirection": self.sort_direction,
            u"srtKey": self.sort_key,
        }
        for key in output_dict:
            yield key, output_dict[key]



from py42.sdk.queries.query_filter import FilterGroup


class BaseQuery(object):
    def __init__(self, *args, **kwargs):
        self._filter_group_list = list(args)
        self._group_clause = kwargs.get(u"group_clause", u"AND")
        self.page_size = 10000
        self.sort_direction = u"asc"

        # Override
        self.page_number = None
        self.sort_key = None

    @classmethod
    def from_dict(cls, _dict, group_clause=u"AND"):
        filter_groups = [FilterGroup.from_dict(item) for item in _dict[u"groups"]]
        return cls(*filter_groups, group_clause=group_clause)

    @classmethod
    def any(cls, *args):
        return cls(*args, group_clause=u"OR")

    @classmethod
    def all(cls, *args):
        return cls(*args)



from py42.services import BaseService


class StoragePreservationDataService(BaseService):

    _base_uri = u"c42api/v3/"

    def __init__(self, main_session, streaming_session):
        super(StoragePreservationDataService, self).__init__(main_session)
        self._streaming_session = streaming_session

    def get_download_token(self, archive_guid, file_id, timestamp):
        """Get PDS download token for a file.

        Args:
            archive_guid (str): Archive guid of the file
            file_id (str): Id of the file.
            timestamp (int): Last updated timestamp of the file in milliseconds.

        Returns:
            :class:`py42.response.Py42Response`: A response containing download token for the file.
        """
        params = {
            u"archiveGuid": archive_guid,
            u"fileId": file_id,
            u"versionTimestamp": timestamp,
        }
        resource = u"FileDownloadToken"
        uri = "{}{}".format(self._base_uri, resource)
        return self._connection.get(uri, params=params)

    def get_file(self, token):
        """Streams a file.

        Args:
            token (str): PDS Download token.

        Returns:
            Returns a stream of the requested token.
        """
        resource = u"GetFile"
        uri = u"{}/{}{}".format(self._connection.host_address, self._base_uri, resource)
        if u"PDSDownloadToken=" in token:
            replaced_token = token.replace(u"PDSDownloadToken=", "")
        else:
            replaced_token = token
        params = {u"PDSDownloadToken": replaced_token}
        headers = {u"Accept": "*/*"}
        return self._streaming_session.get(
            uri, params=params, headers=headers, stream=True
        )



import json

from py42._compat import reprlib
from py42._compat import str
from py42.exceptions import Py42Error


class Py42Response(object):
    def __init__(self, requests_response):
        self._response = requests_response
        self._data = None

    def __getitem__(self, key):
        try:
            return self._data_root[key]
        except TypeError:
            data_root_type = type(self._data_root)
            message = u"The Py42Response root is of type {}, but __getitem__ got a key of {}, which is incompatible.".format(
                data_root_type, key
            )
            raise Py42Error(message)

    def __setitem__(self, key, value):
        try:
            self._data_root[key] = value
        except TypeError:
            data_root_type = type(self._data_root)
            message = u"The Py42Response root is of type {}, but __setitem__ got a key of {} and value of {}, which is incompatible.".format(
                data_root_type, key, value
            )
            raise Py42Error(message)

    def __iter__(self):
        # looping over a Py42Response will loop through list items, dict keys, or str characters
        return iter(self._data_root)

    @property
    def encoding(self):
        """The encoding used to decode the response text."""
        return self._response.encoding

    @property
    def headers(self):
        """A case-insensitive dictionary of response headers."""
        return self._response.headers

    def iter_content(self, chunk_size=1, decode_unicode=False):
        """Iterates over the response data. When ``stream=True`` is set on the request, this avoids
        reading the content at once into memory for large responses.

        Args:
            chunk_size (int, optional): The number of bytes it should read into memory. A value of
                None will function differently depending on the value of `stream`. stream=True will
                read data as it arrives in whatever size the chunks are received. If stream=False,
                data is returned as a single chunk. This is not necessarily the length of each
                item. Defaults to 1.
            decode_unicode (bool, optional): If True, content will be decoded using the best
                available encoding based on the response. Defaults to False.
        """
        return self._response.iter_content(
            chunk_size=chunk_size, decode_unicode=decode_unicode
        )

    @property
    def raw_text(self):
        """The ``response.Response.text`` property. It contains raw metadata that is not included in
        the Py42Response.text property."""
        return self._response.text

    @property
    def text(self):
        """The more useful parts of the HTTP response dumped into a dictionary."""
        return (
            json.dumps(self._data_root)
            if type(self._data_root) != str
            else self._data_root
        )

    @property
    def url(self):
        """The final URL location of response."""
        return self._response.url

    @property
    def status_code(self):
        """An integer code of the response HTTP Status, e.g. 404 or 200."""
        return self._response.status_code

    def __str__(self):
        return str(self._data_root)

    def __repr__(self):
        if not self._response._content_consumed:
            data = "<streamed>"
        else:
            data = self._data_root
        return u"<{} [status={}, data={}]>".format(
            self.__class__.__name__, self._response.status_code, reprlib.repr(data)
        )

    @property
    def content(self):
        return self._response.content

    @property
    def data(self):
        return self._data_root

    @property
    def _data_root(self):
        try:
            if not self._data:
                response_dict = json.loads(self._response.text)
                if type(response_dict) == dict:
                    self._data = response_dict.get(u"data") or response_dict
                else:
                    self._data = response_dict
        except ValueError:
            self._data = self._response.text or u""

        return self._data



from py42._compat import str
from py42.services import BaseService


class FileEventService(BaseService):
    """A service for searching file events.

    See the :ref:`Executing Searches User Guide <anchor_search_file_events>` to learn more about how
    to construct a query.
    """

    def search(self, query):
        """Searches for file events matching the query criteria.
        `REST Documentation <https://forensicsearch-east.us.code42.com/forensic-search/queryservice/swagger-ui.html#/file-event-controller/searchEventsUsingPOST>`__

        Args:
            query (:class:`~py42.sdk.queries.fileevents.file_event_query.FileEventQuery` or str):
                A composed :class:`~py42.sdk.queries.fileevents.file_event_query.FileEventQuery`
                object or the raw query as a JSON formatted string.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the query results.
        """
        query = str(query)
        uri = u"/forensic-search/queryservice/api/v1/fileevent"
        return self._connection.post(uri, data=query)

    def get_file_location_detail_by_sha256(self, checksum):
        """Get file location details based on SHA256 hash.

        Args:
            checksum (str): SHA256 checksum of a file.

        Returns:
            :class:`py42.response.Py42Response`: A response containing file details.
        """
        uri = u"/forensic-search/queryservice/api/v1/filelocations"
        return self._connection.get(uri, params={u"sha256": checksum})



from py42.sdk.queries.query_filter import filter_attributes


class SortDirection(object):
    """Code42 request `sort_direction` constants for sorting returned lists in responses."""

    DESC = u"DESC"
    ASC = u"ASC"

    @staticmethod
    def choices():
        return filter_attributes(SortDirection)



from py42._compat import str


class Py42Error(Exception):
    """A generic, Py42 custom base exception."""


class Py42ResponseError(Py42Error):
    """A base custom class to manage all errors raised because of an HTTP response."""

    def __init__(self, response, message):
        super(Py42ResponseError, self).__init__(message)
        self._response = response

    @property
    def response(self):
        """The response prior to the error."""
        return self._response


class Py42ArchiveFileNotFoundError(Py42ResponseError):
    """An exception raised when a resource file is not found or the path is invalid."""

    def __init__(self, response, device_guid, file_path):
        message = u"File not found in archive for device {} at path {}".format(
            device_guid, file_path
        )
        super(Py42ArchiveFileNotFoundError, self).__init__(response, message)


class Py42ChecksumNotFoundError(Py42ResponseError):
    """An exception raised when a user-supplied hash could not successfully locate its corresponding resource."""

    def __init__(self, response, checksum_name, checksum_value):
        message = u"No files found with {} checksum {}".format(
            checksum_name, checksum_value
        )
        super(Py42ChecksumNotFoundError, self).__init__(response, message)


class Py42FeatureUnavailableError(Py42ResponseError):
    """An exception raised when a requested feature is not supported in your Code42 environment."""

    def __init__(self, response):
        super(Py42FeatureUnavailableError, self).__init__(
            response,
            u"You may be trying to use a feature that is unavailable in your environment.",
        )


class Py42HTTPError(Py42ResponseError):
    """A base custom class to manage all HTTP errors raised by an API endpoint."""

    def __init__(self, exception, message=None):
        message = message or u"Failure in HTTP call {}".format(str(exception))
        super(Py42HTTPError, self).__init__(exception.response, message)


class Py42SecurityPlanConnectionError(Py42HTTPError):
    """An exception raised when the user is not authorized to access the requested resource."""

    def __init__(self, exception, error_message):
        super(Py42SecurityPlanConnectionError, self).__init__(exception, error_message)


class Py42StorageSessionInitializationError(Py42HTTPError):
    """An exception raised when the user is not authorized to initialize a storage session. This
    may occur when trying to restore a file or trying to get events for file activity on removable
    media, in cloud sync folders, and browser uploads."""

    def __init__(self, exception, error_message):
        super(Py42StorageSessionInitializationError, self).__init__(
            exception, error_message
        )


class Py42SessionInitializationError(Py42Error):
    """An exception raised when a user connection is invalid. A connection might be invalid due to
    connection timeout, invalid token, etc.
    """

    def __init__(self, exception):
        error_message = (
            u"An error occurred while requesting "
            u"server environment information, caused by {}".format(str(exception))
        )
        super(Py42SessionInitializationError, self).__init__(exception, error_message)


class Py42BadRequestError(Py42HTTPError):
    """A wrapper to represent an HTTP 400 error."""


class Py42UnauthorizedError(Py42HTTPError):
    """A wrapper to represent an HTTP 401 error."""


class Py42ForbiddenError(Py42HTTPError):
    """A wrapper to represent an HTTP 403 error."""


class Py42NotFoundError(Py42HTTPError):
    """A wrapper to represent an HTTP 404 error."""


class Py42InternalServerError(Py42HTTPError):
    """A wrapper to represent an HTTP 500 error."""


class Py42UserAlreadyAddedError(Py42BadRequestError):
    """An exception raised when the user is already added to group or list, such as the
    Departing Employee list."""

    def __init__(self, exception, user_id, list_name):
        msg = u"User with ID {} is already on the {}.".format(user_id, list_name)
        super(Py42UserAlreadyAddedError, self).__init__(exception, msg)


class Py42LegalHoldNotFoundOrPermissionDeniedError(Py42ForbiddenError):
    """An exception raised when a legal hold matter is inaccessible from your account or
    the matter ID is not valid."""

    def __init__(self, exception, matter_id):
        super(Py42LegalHoldNotFoundOrPermissionDeniedError, self).__init__(
            exception,
            u"Matter with ID={} can not be found. Your account may not have permission to view the matter.".format(
                matter_id
            ),
        )


class Py42InvalidRuleOperationError(Py42HTTPError):
    """An exception raised when trying to add or remove users to a system rule."""

    def __init__(self, exception, rule_id, source):
        msg = u"Only alert rules with a source of 'Alerting' can be targeted by this command. "
        msg += u"Rule {0} has a source of '{1}'."
        super(Py42InvalidRuleOperationError, self).__init__(
            exception, msg.format(rule_id, source)
        )


def raise_py42_error(raised_error):
    """Raises the appropriate :class:`py42.exceptions.Py42HttpError` based on the given
    HTTPError's response status code.
    """
    if raised_error.response.status_code == 400:
        raise Py42BadRequestError(raised_error)
    elif raised_error.response.status_code == 401:
        raise Py42UnauthorizedError(raised_error)
    elif raised_error.response.status_code == 403:
        raise Py42ForbiddenError(raised_error)
    elif raised_error.response.status_code == 404:
        raise Py42NotFoundError(raised_error)
    elif 500 <= raised_error.response.status_code < 600:
        raise Py42InternalServerError(raised_error)
    else:
        raise Py42HTTPError(raised_error)



from py42.sdk.queries.query_filter import create_filter_group
from py42.sdk.queries.query_filter import create_query_filter
from py42.sdk.queries.query_filter import filter_attributes
from py42.sdk.queries.query_filter import QueryFilterStringField
from py42.sdk.queries.query_filter import QueryFilterTimestampField


def create_contains_filter_group(term, value):
    """Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering results
    where the value with key ``term`` contains the given value. Useful for creating ``CONTAINS``
    filters that are not yet supported in py42 or programmatically crafting filter groups.

    Args:
        term: (str): The term of the filter, such as ``actor``.
        value (str): The value used to match on.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """

    filter_list = [create_query_filter(term, u"CONTAINS", value)]
    return create_filter_group(filter_list, u"AND")


def create_not_contains_filter_group(term, value):
    """Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering results
    where the value with key ``term`` does not contain the given value. Useful for creating
    ``DOES_NOT_CONTAIN`` filters that are not yet supported in py42 or programmatically
    crafting filter groups.

    Args:
        term: (str): The term of the filter, such as ``actor``.
        value (str): The value used to exclude on.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """

    filter_list = [create_query_filter(term, u"DOES_NOT_CONTAIN", value)]
    return create_filter_group(filter_list, u"AND")


class AlertQueryFilterStringField(QueryFilterStringField):
    @classmethod
    def contains(cls, value):
        """Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering
        results where the value with key ``self._term`` contains the given value. Useful
        for creating ``CONTAINS`` filters that are not yet supported in py42 or programmatically
        crafting filter groups.

        Args:
            value (str): The value used to match on.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """

        return create_contains_filter_group(cls._term, value)

    @classmethod
    def not_contains(cls, value):
        """Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering
        results where the value with key ``self._term`` does not contain the given value.
        Useful for creating ``DOES_NOT_CONTAIN`` filters that are not yet supported in py42
        or programmatically crafting filter groups.

        Args:
            value (str): The value used to exclude on.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """

        return create_not_contains_filter_group(cls._term, value)


class DateObserved(QueryFilterTimestampField):
    """Class that filters alerts based on the timestamp the alert was triggered."""

    _term = u"createdAt"


class Actor(AlertQueryFilterStringField):
    """Class that filters alerts based on the username that originated the event(s) that
    triggered the alert."""

    _term = u"actor"


class RuleName(AlertQueryFilterStringField):
    """Class that filters alerts based on rule name."""

    _term = u"name"


class RuleId(QueryFilterStringField):
    """Class that filters alerts based on rule identifier."""

    _term = u"ruleId"


class RuleSource(QueryFilterStringField):
    """Class that filters alerts based on rule source.

    Available options are:
        - :attr:`RuleSource.ALERTING`
        - :attr:`RuleSource.DEPARTING_EMPLOYEE`
        - :attr:`RuleSource.HIGH_RISK_EMPLOYEE`
    """

    _term = u"ruleSource"

    ALERTING = u"Alerting"
    DEPARTING_EMPLOYEE = u"Departing Employee"
    HIGH_RISK_EMPLOYEE = u"High Risk Employee"

    @staticmethod
    def choices():
        return filter_attributes(RuleSource)


class RuleType(QueryFilterStringField):
    """Class that filters alerts based on rule type.

    Available options are:
        - :attr:`RuleType.ENDPOINT_EXFILTRATION`
        - :attr:`RuleType.CLOUD_SHARE_PERMISSIONS`
        - :attr:`RuleType.FILE_TYPE_MISMATCH`
    """

    _term = u"type"

    ENDPOINT_EXFILTRATION = u"FedEndpointExfiltration"
    CLOUD_SHARE_PERMISSIONS = u"FedCloudSharePermissions"
    FILE_TYPE_MISMATCH = u"FedFileTypeMismatch"

    @staticmethod
    def choices():
        return filter_attributes(RuleType)


class Description(AlertQueryFilterStringField):
    """Class that filters alerts based on rule description text."""

    _term = u"description"


class Severity(QueryFilterStringField):
    """Class that filters alerts based on severity.

    Available options are:
        - :attr:`Severity.HIGH`
        - :attr:`Severity.MEDIUM`
        - :attr:`Severity.LOW`
    """

    _term = u"severity"

    HIGH = u"HIGH"
    MEDIUM = u"MEDIUM"
    LOW = u"LOW"

    @staticmethod
    def choices():
        return filter_attributes(Severity)


class AlertState(QueryFilterStringField):
    """Class that filters alerts based on alert state.

    Available options are:
        - :attr:`AlertState.OPEN`
        - :attr:`AlertState.DISMISSED`
    """

    _term = u"state"

    OPEN = u"OPEN"
    DISMISSED = u"RESOLVED"

    @staticmethod
    def choices():
        return filter_attributes(AlertState)



from datetime import datetime

from py42.services import BaseService


class StorageSecurityDataService(BaseService):
    def _get_security_detection_events(
        self,
        user_uid=None,
        plan_uid=None,
        cursor=None,
        include_files=None,
        event_types=None,
        min_timestamp=None,
        max_timestamp=None,
        summarize=None,
    ):
        uri = u"/api/SecurityDetectionEvent"

        min_time_str = None
        max_time_str = None

        def get_time_str_from_timestamp(timestamp):
            return datetime.utcfromtimestamp(timestamp).strftime(
                u"%Y-%m-%dT%H:%M:%S.%fZ"
            )

        if min_timestamp:
            min_time_str = get_time_str_from_timestamp(min_timestamp)

        if max_timestamp:
            max_time_str = get_time_str_from_timestamp(max_timestamp)

        params = {
            u"userUid": user_uid,
            u"planUid": plan_uid,
            u"cursor": cursor,
            u"incFiles": include_files,
            u"eventType": event_types,
            u"minTs": min_time_str,
            u"maxTs": max_time_str,
            u"summarize": summarize,
        }

        return self._connection.get(uri, params=params)

    def get_plan_security_events(
        self,
        plan_uid,
        cursor=None,
        include_files=None,
        event_types=None,
        min_timestamp=None,
        max_timestamp=None,
    ):
        return self._get_security_detection_events(
            plan_uid=plan_uid,
            cursor=cursor,
            include_files=include_files,
            event_types=event_types,
            min_timestamp=min_timestamp,
            max_timestamp=max_timestamp,
        )

    def get_user_security_events(
        self,
        user_uid,
        cursor=None,
        include_files=None,
        event_types=None,
        min_timestamp=None,
        max_timestamp=None,
    ):
        return self._get_security_detection_events(
            user_uid=user_uid,
            cursor=cursor,
            include_files=include_files,
            event_types=event_types,
            min_timestamp=min_timestamp,
            max_timestamp=max_timestamp,
        )

    def get_security_detection_event_summary(
        self, user_uid, cursor=None, min_timestamp=None, max_timestamp=None
    ):
        return self._get_security_detection_events(
            user_uid=user_uid,
            cursor=cursor,
            min_timestamp=min_timestamp,
            max_timestamp=max_timestamp,
            summarize=True,
        )



from collections import namedtuple


class BaseService(object):

    __slots__ = ["_connection"]

    def __init__(self, connection):
        self._connection = connection


Services = namedtuple(
    u"Services",
    [
        u"administration",
        u"archive",
        u"devices",
        u"legalhold",
        u"orgs",
        u"securitydata",
        u"users",
        u"alertrules",
        u"alerts",
        u"fileevents",
        u"savedsearch",
        u"preservationdata",
        u"departingemployee",
        u"highriskemployee",
        u"userprofile",
    ],
)



from __future__ import print_function

import json as json_lib
from threading import Lock

from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from requests.models import Request
from requests.sessions import Session

import py42.settings as settings
from py42._compat import urljoin
from py42._compat import urlparse
from py42.exceptions import Py42Error
from py42.exceptions import Py42FeatureUnavailableError
from py42.exceptions import raise_py42_error
from py42.response import Py42Response
from py42.services._auth import C42RenewableAuth
from py42.settings import debug
from py42.util import format_dict

SESSION_ADAPTER = HTTPAdapter(pool_connections=200, pool_maxsize=4, pool_block=True)

ROOT_SESSION = Session()
ROOT_SESSION.mount(u"https://", SESSION_ADAPTER)
ROOT_SESSION.mount(u"http://", SESSION_ADAPTER)
ROOT_SESSION.headers = {
    u"Accept": u"application/json",
    u"Content-Type": u"application/json",
    u"Accept-Encoding": u"gzip, deflate",
    u"Connection": u"keep-alive",
}


class HostResolver(object):
    def get_host_address(self):
        raise NotImplementedError()


class KnownUrlHostResolver(HostResolver):
    def __init__(self, host_address):
        self._host_address = host_address

    def get_host_address(self):
        return self._host_address


class MicroservicePrefixHostResolver(HostResolver):
    def __init__(self, connection, prefix):
        self._connection = connection
        self._prefix = prefix

    def get_host_address(self):
        sts_url = self._get_sts_base_url()
        return sts_url.replace(u"sts", self._prefix, 1)

    def _get_sts_base_url(self):
        uri = u"/api/ServerEnv"
        response = self._connection.get(uri)

        response_json = json_lib.loads(response.text)
        sts_base_url = response_json.get(u"stsBaseUrl")

        if not sts_base_url:
            raise Py42FeatureUnavailableError(response)

        return sts_base_url


class MicroserviceKeyHostResolver(HostResolver):
    def __init__(self, kv_service, key):
        self._kv_service = kv_service
        self._key = key

    def get_host_address(self):
        return self._kv_service.get_stored_value(self._key).text


class Connection(object):
    def __init__(self, host_resolver, auth=None, session=None):
        self._host_resolver = host_resolver
        self._session = session or ROOT_SESSION
        self._headers = self._session.headers.copy()
        self._auth = auth
        self._resolve_lock = Lock()
        self._host_address = None

    @classmethod
    def from_host_address(cls, host_address, auth=None, session=None):
        host_resolver = KnownUrlHostResolver(host_address)
        return cls(host_resolver, auth=auth, session=session)

    @classmethod
    def from_microservice_key(cls, kv_service, key, auth=None, session=None):
        host_resolver = MicroserviceKeyHostResolver(kv_service, key)
        return cls(host_resolver, auth=auth, session=session)

    @classmethod
    def from_microservice_prefix(cls, connection, prefix, auth=None, session=None):
        host_resolver = MicroservicePrefixHostResolver(connection, prefix)
        return cls(host_resolver, auth=auth, session=session)

    @property
    def host_address(self):
        return self._get_host_address()

    def clone(self, host_address):
        host_resolver = KnownUrlHostResolver(host_address)
        return Connection(host_resolver, auth=self._auth)

    def get(self, url, **kwargs):
        return self.request(u"GET", url, **kwargs)

    def options(self, url, **kwargs):
        return self.request(u"OPTIONS", url, **kwargs)

    def head(self, url, **kwargs):
        return self.request(u"HEAD", url, **kwargs)

    def post(self, url, data=None, **kwargs):
        return self.request(u"POST", url, data=data, **kwargs)

    def put(self, url, data=None, **kwargs):
        return self.request(u"PUT", url, data=data, **kwargs)

    def patch(self, url, data=None, **kwargs):
        return self.request(u"PATCH", url, data=data, **kwargs)

    def delete(self, url, **kwargs):
        return self.request(u"DELETE", url, **kwargs)

    def request(
        self,
        method,
        url,
        params=None,
        data=None,
        json=None,
        headers=None,
        cookies=None,
        files=None,
        auth=None,
        hooks=None,
        stream=False,
        timeout=60,
        cert=None,
        proxies=None,
    ):
        response = None
        for _ in range(2):
            request = self._prepare_request(
                method,
                url,
                params=params,
                data=data,
                json=json,
                headers=headers,
                cookies=cookies,
                files=files,
                auth=auth,
                hooks=hooks,
            )

            response = self._session.send(
                request,
                stream=stream,
                timeout=timeout,
                verify=settings.verify_ssl_certs,
                cert=cert,
                proxies=proxies,
            )

            if not stream and response is not None:
                # setting this manually speeds up read times
                response.encoding = u"utf-8"

            if response is not None and 200 <= response.status_code <= 399:
                return Py42Response(response)

            if response is not None and response.status_code == 401:
                if isinstance(self._auth, C42RenewableAuth):
                    self._auth.clear_credentials()

        # if nothing has been returned after two attempts, something went wrong
        _handle_error(method, url, response)

    def _prepare_request(
        self,
        method,
        url,
        params=None,
        data=None,
        json=None,
        headers=None,
        cookies=None,
        files=None,
        auth=None,
        hooks=None,
    ):
        url = urljoin(self.host_address, url)
        self._session.proxies = settings.proxies
        self._session.verify = settings.verify_ssl_certs

        if json is not None:
            data = json_lib.dumps(json)

        user_headers = {u"User-Agent": settings.get_user_agent_string()}
        if headers:
            user_headers.update(headers)
        self._headers.update(user_headers)
        request = Request(
            method=method,
            url=url,
            headers=self._headers,
            files=files,
            data=data,
            params=params,
            auth=auth or self._auth,
            cookies=cookies,
            hooks=hooks,
        )

        _print_request(method, url, params=params, data=data)
        return self._session.prepare_request(request)

    def _get_host_address(self):
        if not self._host_address:
            with self._resolve_lock:
                if not self._host_address:
                    host = self._host_resolver.get_host_address()
                    self._init_host_info(host)
        return self._host_address

    def _init_host_info(self, host):
        if not host.startswith(u"http://") and not host.startswith(u"https://"):
            host = u"https://{}".format(host)
        parsed_host = urlparse(host)
        self._headers[u"Host"] = parsed_host.netloc
        self._host_address = host


def _handle_error(method, url, response):
    if response is None:
        msg = u"No response was returned for {} request to {}.".format(method, url)
        raise Py42Error(msg)

    try:
        response.raise_for_status()
    except HTTPError as ex:
        raise_py42_error(ex)


def _print_request(method, url, params=None, data=None):
    debug.logger.info(u"{}{}".format(method.ljust(8), url))
    if params:
        debug.logger.debug(format_dict(params, u"  params"))
    if data:
        debug.logger.debug(format_dict(data, u"  data"))



from py42 import settings
from py42.services import BaseService
from py42.services.util import get_all_pages


class ArchiveService(BaseService):
    def get_data_key_token(self, device_guid):
        uri = u"/api/DataKeyToken"
        data = {u"computerGuid": device_guid}
        return self._connection.post(uri, json=data)

    def get_single_archive(self, archive_guid):
        """Gets single archive information by GUID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Archive-get>`__

        Args:
            archive_guid (str): The GUID for the archive.

        Returns:
            :class:`py42.response.Py42Response`: A response containing archive information.
        """
        uri = u"/api/Archive/{}".format(archive_guid)
        return self._connection.get(uri)

    def get_page(self, page_num, page_size=None, **kwargs):
        """Gets an individual page of archives.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Archive-get>`__

        Args:
            page_num (int): The page number to request.
            page_size (int, optional): The number of archives to return per page. Defaults to `py42.settings.items_per_page`.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/Archive"
        page_size = page_size or settings.items_per_page
        params = dict(pgNum=page_num, pgSize=page_size, **kwargs)
        return self._connection.get(uri, params=params)

    def get_all_archives_from_value(self, id_value, id_type):
        """Gets archive information from an ID, such as a User UID, Device GUID, or Destination GUID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Archive-get>`__

        Args:
            id_value (str): Query value for archive.
            id_type (str): API query value description (e.g backupSourceGuid,
                userUid, destinationGuid)

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of archives.
        """
        params = {u"{}".format(id_type): u"{}".format(id_value)}
        return get_all_pages(self.get_page, u"archives", **params)

    def get_backup_sets(self, device_guid, destination_guid):
        uri = u"/c42api/v3/BackupSets/{}/{}".format(device_guid, destination_guid)
        return self._connection.get(uri)

    def get_all_restore_history(self, days, id_type, id_value, **kwargs):
        return get_all_pages(
            self._get_restore_history_page,
            u"restoreEvents",
            days=days,
            id_type=id_type,
            id_value=id_value,
            **kwargs
        )

    def _get_restore_history_page(
        self, days, id_type, id_value, page_num, page_size, **kwargs
    ):
        uri = u"/api/RestoreHistory"
        params = dict(days=days, pgNum=page_num, pgSize=page_size, **kwargs)
        params[id_type] = id_value
        return self._connection.get(uri, params=params)

    def get_web_restore_info(self, src_guid, dest_guid):
        uri = u"/api/WebRestoreInfo"
        params = {u"srcGuid": src_guid, u"destGuid": dest_guid}
        return self._connection.get(uri, params=params)

    def _get_cold_storage_archives_page(
        self,
        org_id=None,
        include_child_orgs=None,
        sort_key=u"archiveHoldExpireDate",
        sort_dir=u"asc",
        page_size=None,
        page_num=None,
    ):
        uri = u"/api/ColdStorage"
        params = {
            u"orgId": org_id,
            u"incChildOrgs": include_child_orgs,
            u"srtKey": sort_key,
            u"srtDir": sort_dir,
            u"pgSize": page_size,
            u"pgNum": page_num,
        }

        return self._connection.get(uri, params=params)

    def get_all_org_cold_storage_archives(
        self,
        org_id,
        include_child_orgs=True,
        sort_key=u"archiveHoldExpireDate",
        sort_dir=u"asc",
    ):
        return get_all_pages(
            self._get_cold_storage_archives_page,
            u"coldStorageRows",
            org_id=org_id,
            include_child_orgs=include_child_orgs,
            sort_key=sort_key,
            sort_dir=sort_dir,
        )

    def update_cold_storage_purge_date(self, archive_guid, purge_date):
        uri = u"/api/coldStorage/{}".format(archive_guid)
        params = {u"idType": u"guid"}
        data = {u"archiveHoldExpireDate": purge_date}
        return self._connection.put(uri, params=params, json=data)



from py42.sdk.queries.fileevents.file_event_query import FileEventFilterStringField


class DeviceUsername(FileEventFilterStringField):
    """Class that filters events by the Code42 username of the device that observed the event."""

    _term = u"deviceUserName"


class OSHostname(FileEventFilterStringField):
    """Class that filters events by hostname of the device that observed the event."""

    _term = u"osHostName"


class PrivateIPAddress(FileEventFilterStringField):
    """Class that filters events by private (LAN) IP address of the device that observed the event."""

    _term = u"privateIpAddresses"


class PublicIPAddress(FileEventFilterStringField):
    """Class that filters events by public (WAN) IP address of the device that observed the event."""

    _term = u"publicIpAddress"


class DeviceSignedInUserName(FileEventFilterStringField):
    """Class that filters events by signed in user of the device that observed the event."""

    _term = u"operatingSystemUser"



class AlertsClient(object):
    def __init__(self, alert_service, alert_rules_client):
        self._alert_service = alert_service
        self._alert_rules_client = alert_rules_client

    @property
    def rules(self):
        """A collection of methods for managing alert rules.

        Returns:
            :class:`py42.services.alertrules.AlertRulesClient`
        """
        return self._alert_rules_client

    def search(self, query):
        """Searches alerts using the given :class:`py42.sdk.queries.alerts.alert_query.AlertQuery`.

        Args:
            query (:class:`py42.sdk.queries.alerts.alert_query.AlertQuery`): An alert query.
                See the :ref:`Executing Searches User Guide <anchor_search_alerts>` to learn more
                about how to construct a query.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the alerts that match the given
            query.
        """
        return self._alert_service.search(query)

    def get_details(self, alert_ids):
        """Gets the details for the alerts with the given IDs, including the file event query that,
        when passed into a search, would result in events that could have triggered the alerts.

        Args:
            alert_ids (iter[str]): The identification numbers of the alerts for which you want to
                get details for.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the alert details.
        """
        return self._alert_service.get_details(alert_ids)

    def resolve(self, alert_ids, reason=None):
        """Resolves the alerts with the given IDs.

        Args:
            alert_ids (iter[str]): The identification numbers for the alerts to resolve.
            reason (str, optional): The reason the alerts are now resolved. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._alert_service.resolve(alert_ids, reason=reason)

    def reopen(self, alert_ids, reason=None):
        """Reopens the resolved alerts with the given IDs.

        Args:
            alert_ids (iter[str]): The identification numbers for the alerts to reopen.
            reason (str, optional): The reason the alerts are reopened. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._alert_service.reopen(alert_ids, reason=reason)



"""
This module handles import compatibility issues between Python 2 and
Python 3.
"""
import sys

_ver = sys.version_info

#: Python 2.x?
is_py2 = _ver[0] == 2

if is_py2:
    from urllib import quote
    from urllib import urlencode

    from urlparse import urljoin
    from urlparse import urlparse

    str = unicode

    import repr as reprlib

    string_type = basestring

else:
    from urllib.parse import urljoin
    from urllib.parse import urlparse
    from urllib.parse import quote
    from urllib.parse import urlencode

    str = str

    import reprlib

    string_type = str



from collections import namedtuple

AuthorityClient = namedtuple(
    u"AuthorityClient",
    [
        u"administration",
        u"archive",
        u"devices",
        u"legalhold",
        u"orgs",
        u"securitydata",
        u"users",
    ],
)



from py42.sdk.queries.fileevents.file_event_query import FileEventQuery
from py42.services import BaseService


class SavedSearchService(BaseService):
    """A service to interact with saved search APIs."""

    _version = u"v1"
    _resource = u"/forensic-search/queryservice/api/{}/saved".format(_version)

    def __init__(self, connection, file_event_client):
        super(SavedSearchService, self).__init__(connection)
        self._file_event_client = file_event_client

    def get(self):
        """Fetch details of existing saved searches.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"{}".format(self._resource)
        return self._connection.get(uri)

    def get_by_id(self, search_id):
        """Fetch the details of a saved search by its given search Id.

        Args:
            search_id (str): Unique search Id of the saved search.
        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"{}/{}".format(self._resource, search_id)
        return self._connection.get(uri)

    def get_query(self, search_id):
        """Get the saved search in form of a query(`py42.sdk.queries.fileevents.file_event_query`).

        Args:
            search_id (str): Unique search Id of the saved search.
        Returns:
            :class:`py42.sdk.queries.fileevents.file_event_query.FileEventQuery`
        """
        response = self.get_by_id(search_id)
        search = response[u"searches"][0]
        return FileEventQuery.from_dict(search)

    def execute(self, search_id, pg_num=1, pg_size=10000):
        """
        Execute a saved search for given search Id and return its results.

        Args:
            search_id (str): Unique search Id of the saved search.
            pg_num (int, optional): The consecutive group of results of size pg_size in the result set to return. Defaults to 1.
            pg_size (int, optional): The maximum number of results to be returned. Defaults to 10,000.
        Returns:
            :class:`py42.response.Py42Response`
        """
        query = self.get_query(search_id)
        return self._file_event_client.search(query)



from py42.exceptions import Py42UserAlreadyAddedError

_PAGE_SIZE = 100


def handle_user_already_added_error(bad_request_err, username_tried_adding, list_name):
    if u"User already on list" in bad_request_err.response.text:
        raise Py42UserAlreadyAddedError(
            bad_request_err, username_tried_adding, list_name
        )


class _DetectionListFilters(object):
    OPEN = u"OPEN"
    EXFILTRATION_30_DAYS = u"EXFILTRATION_30_DAYS"
    EXFILTRATION_24_HOURS = u"EXFILTRATION_24_HOURS"



from collections import OrderedDict
from datetime import datetime

from py42._compat import str
from py42._compat import string_type
from py42.util import convert_datetime_to_timestamp_str
from py42.util import convert_timestamp_to_str


def create_query_filter(term, operator, value=None):
    """Creates a :class:`~py42.sdk.queries.query_filter.QueryFilter` object. Useful for
    programmatically crafting query filters, such as filters not yet defined in py42.

    Args:
        term (str): The term of the filter, such as ``actor`` or ``sharedWith``.
        operator (str): The operator between ``term`` and ``value``, such as ``IS`` or `IS_NOT`.
        value (str): The value used to filter results.

    Returns:
        :class:`~py42.sdk.queries.query_filter.QueryFilter`
    """

    return QueryFilter(term, operator, value)


def create_filter_group(query_filter_list, filter_clause):
    """Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` object. Useful for
    programmatically crafting query filters, such as filters not yet defined in py42.
    Alternatively, if you want to create custom filter groups with already defined
    operators (such as `IS` or `IS_IN`), see the other methods in this module, such as
    :meth:`~py42.sdk.queries.query_filter.create_eq_filter_group()`.

    Args:
        query_filter_list (list): a list of :class:`~py42.sdk.queries.query_filter.QueryFilter`
            objects.
        filter_clause (str): The clause joining the filters, such as ``AND`` or ``OR``.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """

    return FilterGroup(query_filter_list, filter_clause)


def create_eq_filter_group(term, value):
    """"Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering results
    where the value with key ``term`` equals the given value. Useful for creating ``IS``
    filters that are not yet supported in py42 or programmatically crafting filter groups.

    Args:
        term: (str): The term of the filter, such as ``actor`` or ``sharedWith``.
        value (str): The value used to match on.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """

    filter_list = [create_query_filter(term, u"IS", value)]
    return create_filter_group(filter_list, u"AND")


def create_not_eq_filter_group(term, value):
    """"Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering results
    where the value with key ``term`` does not equal the given value. Useful for creating
    ``IS_NOT`` filters that are not yet supported in py42 or programmatically crafting filter
    groups.

    Args:
        term: (str): The term of the filter, such as ``actor`` or ``sharedWith``.
        value (str): The value used to exclude on.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """

    filter_list = [create_query_filter(term, u"IS_NOT", value)]
    return create_filter_group(filter_list, u"AND")


def create_is_in_filter_group(term, value_list):
    """"Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering results
    where the value with key ``term`` is one of several values. Useful for creating ``IS_IN``
    filters that are not yet supported in py42 or programmatically crafting filter groups.

    Args:
        term: (str): The term of the filter, such as ``actor`` or ``sharedWith``.
        value_list (list): The list of values to match on.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """

    filter_list = [create_query_filter(term, u"IS", value) for value in value_list]
    return create_filter_group(filter_list, u"OR" if len(filter_list) > 1 else u"AND")


def create_not_in_filter_group(term, value_list):
    """"Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering results
    where the value with key ``term`` is not one of several values. Useful for creating
    ``NOT_IN`` filters that are not yet supported in py42 or programmatically crafting
    filter groups.

    Args:
        term: (str): The term of the filter, such as ``actor`` or ``sharedWith``.
        value_list (list): The list of values to exclude on.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """

    filter_list = [create_query_filter(term, u"IS_NOT", value) for value in value_list]
    return create_filter_group(filter_list, u"AND")


def create_on_or_after_filter_group(term, value):
    """"Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering results
    where the value with key ``term`` is on or after the given value. Examples include
    values describing dates. Useful for creating ``ON_OR_AFTER`` filters that are not yet
    supported in py42  or programmatically crafting filter groups.

    Args:
        term: (str): The term of the filter, such as ``eventTimestamp``.
        value (str or int): The value used to filter results.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """

    filter_list = [create_query_filter(term, u"ON_OR_AFTER", value)]
    return create_filter_group(filter_list, u"AND")


def create_on_or_before_filter_group(term, value):
    """"Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering results
    where the value with key ``term`` is on or before the given value. Examples include
    values describing dates. Useful for creating ``ON_OR_BEFORE`` filters that are not
    yet supported in py42 or programmatically crafting filter groups.

    Args:
        term: (str): The term of the filter, such as ``eventTimestamp``.
        value (str or int): The value used to filter results.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """

    filter_list = [create_query_filter(term, u"ON_OR_BEFORE", value)]
    return create_filter_group(filter_list, u"AND")


def create_in_range_filter_group(term, start_value, end_value):
    """"Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for filtering results
    where the value with key ``term`` is in the given range. Examples include values describing
    dates. Useful for creating a combination of ``ON_OR_AFTER`` and ``ON_OR_BEFORE`` filters
    that are not yet supported in py42 or programmatically crafting filter groups.

    Args:
        term: (str): The term of the filter, such as ``eventTimestamp``.
        start_value (str or int): The start value used to filter results.
        end_value (str or int): The end value used to filter results.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """

    filter_list = [
        create_query_filter(term, u"ON_OR_AFTER", start_value),
        create_query_filter(term, u"ON_OR_BEFORE", end_value),
    ]
    return create_filter_group(filter_list, u"AND")


def create_within_the_last_filter_group(term, value):
    """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
    for finding results where the key ``term`` is an ``EventTimestamp._term``
    and the value is one of the `EventTimestamp` attributes as `value`.

    Args:
        value (str): `EventTimestamp` attribute.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """
    filter_list = [create_query_filter(term, u"WITHIN_THE_LAST", value)]
    return create_filter_group(filter_list, u"AND")


def filter_attributes(cls):

    return [
        cls().__getattribute__(attr)
        for attr in dir(cls)
        if not callable(cls().__getattribute__(attr)) and not attr.startswith(u"_")
    ]


class QueryFilterStringField(object):
    """Helper class for creating filters where the search value is a string."""

    _term = u"override_string_field_name"

    @classmethod
    def eq(cls, value):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the value with key ``self._term`` equals the provided
        ``value``.

        Args:
            value (str): The value to match on.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        return create_eq_filter_group(cls._term, value)

    @classmethod
    def not_eq(cls, value):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the value with key ``self._term`` does not equal the provided ``value``.

        Args:
            value (str): The value to exclude on.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        return create_not_eq_filter_group(cls._term, value)

    @classmethod
    def is_in(cls, value_list):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the value with the key ``self._term`` is in the provided
        ``value_list``.

        Args:
            value_list (list): The list of values to match on.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        return create_is_in_filter_group(cls._term, value_list)

    @classmethod
    def not_in(cls, value_list):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the value with the key ``self._term`` is not in the provided
        ``value_list``.

        Args:
            value_list (list): The list of values to exclude on.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        return create_not_in_filter_group(cls._term, value_list)


class QueryFilterTimestampField(object):
    """Helper class for creating filters where the search value is a timestamp."""

    _term = u"override_timestamp_field_name"

    @classmethod
    def on_or_after(cls, value):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the value with key ``self._term` is on or after the
        provided ``value``.

        Args:
            value (str or int): The value used to filter results.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        formatted_timestamp = convert_timestamp_to_str(value)
        return create_on_or_after_filter_group(cls._term, formatted_timestamp)

    @classmethod
    def on_or_before(cls, value):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the value with key ``self._term`` is on or before the
        provided ``value``.

        Args:
            value (str or int): The value used to filter results.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        formatted_timestamp = convert_timestamp_to_str(value)
        return create_on_or_before_filter_group(cls._term, formatted_timestamp)

    @classmethod
    def in_range(cls, start_value, end_value):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the value with key ``self._term`` is in range between
        the provided ``start_value`` and ``end_value``.

        Args:
            start_value (str or int): The start value used to filter results.
            end_value (str or int): The end value used to filter results.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        formatted_start_time = convert_timestamp_to_str(start_value)
        formatted_end_time = convert_timestamp_to_str(end_value)
        return create_in_range_filter_group(
            cls._term, formatted_start_time, formatted_end_time
        )

    @classmethod
    def on_same_day(cls, value):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the value with key ``self._term`` is within the same
        calendar day as the provided ``value``.

        Args:
            value (str or int): The value used to filter results.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        date_from_value = datetime.utcfromtimestamp(value)
        start_time = datetime(
            date_from_value.year, date_from_value.month, date_from_value.day, 0, 0, 0
        )
        end_time = datetime(
            date_from_value.year, date_from_value.month, date_from_value.day, 23, 59, 59
        )
        formatted_start_time = convert_datetime_to_timestamp_str(start_time)
        formatted_end_time = convert_datetime_to_timestamp_str(end_time)
        return create_in_range_filter_group(
            cls._term, formatted_start_time, formatted_end_time
        )

    @classmethod
    def within_the_last(cls, value):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the key ``self._term`` is an ``EventTimestamp._term``
        and the value is one of the ``EventTimestamp`` attributes as ``value``.

        Args:
            value (str): `EventTimestamp` attribute.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        return create_within_the_last_filter_group(cls._term, value)


class QueryFilterBooleanField(object):
    """Helper class for creating filters where the search value is a boolean."""

    _term = u"override_boolean_field_name"

    @classmethod
    def is_true(cls):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the value with key ``self._term`` is True.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        return create_eq_filter_group(cls._term, u"TRUE")

    @classmethod
    def is_false(cls):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the value with key ``self._term`` is False.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        return create_eq_filter_group(cls._term, u"FALSE")


class QueryFilter(object):
    """Class for constructing a single filter object for use in a search query.

    When :func:`str()` is called on a :class:`~py42.sdk.queries.query_filter.QueryFilter`
    instance, the (``term``, ``operator``, ``value``) attribute combination is transformed
    into a JSON string to be used as part of a Forensic Search or Alert query.

    When :func:`dict()` is called on a :class:`~py42.sdk.queries.query_filter.QueryFilter`
    instance, the (``term``, ``operator``, ``value``) attribute combination is transformed
    into the Python `dict` equivalent of their JSON representation. This can be useful
    for programmatically manipulating a :class:`~py42.sdk.queries.query_filter.QueryFilter`
    after it's been created.
    """

    _term = None

    def __init__(self, term, operator, value=None):
        self._term = term
        self._operator = operator
        self._value = value

    @classmethod
    def from_dict(cls, _dict):
        """Creates an instance of :class:`~py42.sdk.queries.query_filter.QueryFilter` from
        the values found in ``_dict``. ``_dict`` must contain keys ``term``, ``operator``,
        and ``value``.

        Args:
            _dict (dict): A dictionary containing keys ``term``, ``operator``, and ``value``.

        Returns:
            :class:`~py42.sdk.queries.query_filter.QueryFilter`
        """

        return cls(_dict[u"term"], _dict[u"operator"], value=_dict.get(u"value"))

    @property
    def term(self):
        """The term of the filter, such as ``actor`` or ``sharedWith``."""

        return self._term

    @property
    def operator(self):
        """The operator between ``term`` and ``value``, such as ``IS`` or `IS_NOT`."""

        return self._operator

    @property
    def value(self):
        """The value used to filter results."""

        return self._value

    def __str__(self):
        value = u"null" if self._value is None else u'"{}"'.format(self._value)
        return u'{{"operator":"{0}", "term":"{1}", "value":{2}}}'.format(
            self._operator, self._term, value
        )

    def __iter__(self):
        output_dict = OrderedDict()
        output_dict[u"operator"] = self._operator
        output_dict[u"term"] = self._term
        output_dict[u"value"] = self._value
        for key in output_dict:
            yield key, output_dict[key]

    def __eq__(self, other):
        if isinstance(other, (QueryFilter, tuple, list)):
            return tuple(self) == tuple(other)
        elif isinstance(other, string_type):
            return str(self) == other
        else:
            return False

    def __hash__(self):
        return hash(str(self))


class FilterGroup(object):
    """Class for constructing a logical sub-group of related filters from a list of
    :class:`~py42.sdk.queries.query_filter.QueryFilter` objects. Takes a list of
    :class:`~py42.sdk.queries.query_filter.QueryFilter` objects and combines them
    logically using the passed in filter clause (``AND`` or ``OR``).

    When :func:`str()` is called on a :class:`FilterGroup` instance, the combined filter items are
    transformed into a JSON string to be used as part of a Forensic Search or Alert query.

    When :func:`dict()` is called on a :class:`~py42.sdk.queries.query_filter.FilterGroup`
    instance, the combined filter items are transformed into the Python `dict` equivalent
    of their JSON representation. This can be useful for programmatically manipulating a
    :class:`~py42.sdk.queries.query_filter.FilterGroup` after it's been created.
    """

    def __init__(self, filter_list, filter_clause=u"AND"):
        self._filter_list = filter_list
        self._filter_clause = filter_clause

    @classmethod
    def from_dict(cls, _dict):
        """Creates an instance of :class:`~py42.sdk.queries.query_filter.FilterGroup`
        from the values found in ``_dict``. ``_dict`` must contain keys ``filters`` and
        ``filterClause``.

        Args:
            _dict (dict): A dictionary containing keys ``term``, ``operator``, and ``value``.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        filter_list = [QueryFilter.from_dict(item) for item in _dict[u"filters"]]
        return cls(filter_list, filter_clause=_dict[u"filterClause"])

    @property
    def filter_list(self):
        """The list of :class:`~py42.sdk.queries.query_filter.QueryFilter` objects in this
        group."""

        return self._filter_list

    @property
    def filter_clause(self):
        """The clause joining the filters, such as ``AND`` or ``OR``."""

        return self._filter_clause

    @filter_clause.setter
    def filter_clause(self, value):
        """The clause joining the filters, such as ``AND`` or ``OR``."""

        self._filter_clause = value

    @property
    def _filter_set(self):
        return sorted(list(set(self.filter_list)), key=str)

    def __str__(self):
        filters_string = u",".join(str(filter_item) for filter_item in self._filter_set)
        return u'{{"filterClause":"{0}", "filters":[{1}]}}'.format(
            self._filter_clause, filters_string
        )

    def __iter__(self):
        filter_list = [dict(item) for item in self._filter_set]
        output_dict = {u"filterClause": self._filter_clause, u"filters": filter_list}
        for key in output_dict:
            yield key, output_dict[key]

    def __eq__(self, other):
        if isinstance(other, FilterGroup):
            return (
                self.filter_clause == other.filter_clause
                and self._filter_set == other._filter_set
            )
        elif isinstance(other, (tuple, list)):
            return tuple(self) == tuple(other)
        elif isinstance(other, string_type):
            return str(self) == other
        else:
            return False

    def __contains__(self, item):
        return item in self._filter_set



import logging
import sys


class _DebugSettings(object):
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    TRACE = logging.DEBUG
    NONE = logging.NOTSET

    def __init__(self):
        self.logger = logging.getLogger("py42")
        self.logger.addHandler(logging.StreamHandler(sys.stderr))

    @property
    def level(self):
        return self.logger.level

    @level.setter
    def level(self, level):
        self.logger.setLevel(level)


sys.modules[__name__] = _DebugSettings()



from py42 import settings
from py42._compat import quote
from py42.services import BaseService
from py42.services.util import get_all_pages


class UserService(BaseService):
    """A service for interacting with Code42 user APIs. Use the UserService to create and retrieve
    users. You can also use it to block and deactivate users.
    """

    def create_user(
        self,
        org_uid,
        username,
        email,
        password=None,
        first_name=None,
        last_name=None,
        notes=None,
    ):
        """Creates a new user.
        WARNING: If the provided username already exists for a user, it will be updated in the
        database instead.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#User-post>`__

        Args:
            org_uid (str): The org UID for the organization the new user belongs to.
            username (str): The username for the new user.
            email (str): The email for the new user.
            password (str, optional): The password for the new user. Defaults to None.
            first_name (str, optional): The first name for the new user. Defaults to None.
            last_name (str, optional): The last name for the new user. Defaults to None.
            notes (str, optional): Descriptive information about the user. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """

        uri = u"/api/User"
        data = {
            u"orgUid": org_uid,
            u"username": username,
            u"email": email,
            u"password": password,
            u"firstName": first_name,
            u"lastName": last_name,
            u"notes": notes,
        }
        return self._connection.post(uri, json=data)

    def get_by_id(self, user_id, **kwargs):
        """Gets the user with the given ID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#User-get>`__

        Args:
            user_id (int): An ID for a user.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the user.
        """
        uri = u"/api/User/{}".format(user_id)
        return self._connection.get(uri, params=kwargs)

    def get_by_uid(self, user_uid, **kwargs):
        """Gets the user with the given UID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#User-get>`__

        Args:
            user_uid (str): A UID for a user.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the user.
        """
        uri = u"/api/User/{}".format(user_uid)
        params = dict(idType=u"uid", **kwargs)
        return self._connection.get(uri, params=params)

    def get_by_username(self, username, **kwargs):
        """Gets the user with the given username.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#User-get>`__

        Args:
            username (str): A username for a user.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the user.
        """
        uri = u"/api/User"
        params = dict(username=username, **kwargs)
        return self._connection.get(uri, params=params)

    def get_current(self, **kwargs):
        """Gets the currently signed in user.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#User-get>`__

        Returns:
            :class:`py42.response.Py42Response`: A response containing the user.
        """
        uri = u"/api/User/my"
        return self._connection.get(uri, params=kwargs)

    def get_page(
        self,
        page_num,
        active=None,
        email=None,
        org_uid=None,
        role_id=None,
        page_size=None,
        q=None,
        **kwargs
    ):
        """Gets an individual page of users.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#User-get>`__

        Args:
            page_num (int): The page number to request.
            active (bool, optional): True gets active users only,
                and false gets deactivated users only. Defaults to None.
            email (str, optional): Limits users to only those with this email. Defaults to None.
            org_uid (str, optional): Limits users to only those in the organization with this org
                UID. Defaults to None.
            role_id (int, optional): Limits users to only those with a given role ID. Defaults to
                None.
            page_size (int, optional): The number of items on the page. Defaults to `py42.settings.items_per_page`.
            q (str, optional): A generic query filter that searches across name, username, and
                email. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """

        uri = u"/api/User"
        page_size = page_size or settings.items_per_page
        params = dict(
            active=active,
            email=email,
            orgUid=org_uid,
            roleId=role_id,
            pgNum=page_num,
            pgSize=page_size,
            q=q,
            **kwargs
        )

        return self._connection.get(uri, params=params)

    def get_all(
        self, active=None, email=None, org_uid=None, role_id=None, q=None, **kwargs
    ):
        """Gets all users.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#User-get>`__

        Args:
            active (bool, optional): True gets active users only,
                and false gets deactivated users only. Defaults to None.
            email (str, optional): Limits users to only those with this email. Defaults to None.
            org_uid (str, optional): Limits users to only those in the organization with this org
                UID. Defaults to None.
            role_id (int, optional): Limits users to only those with a given role ID. Defaults to
                None.
            q (str, optional): A generic query filter that searches across name, username, and
                email. Defaults to None.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of users.
        """
        return get_all_pages(
            self.get_page,
            u"users",
            active=active,
            email=email,
            org_uid=org_uid,
            role_id=role_id,
            q=q,
            **kwargs
        )

    def get_scim_data_by_uid(self, user_uid):
        """Returns SCIM data such as division, department, and title for a given user.
        `REST Documentation <https://console.us.code42.com/swagger/#/scim-user-data/ScimUserData_CollatedView>`__

        Args:
            user_uid (str): A Code42 user uid.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/v7/scim-user-data/collated-view"
        params = dict(userId=user_uid)
        return self._connection.get(uri, params=params)

    def block(self, user_id):
        """Blocks the user with the given ID. A blocked user is not allowed to log in or restore
        files. Backups will continue if the user is still active.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#UserBlock-put>`__

        Args:
            user_id (int): An ID for a user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/UserBlock/{}".format(user_id)
        return self._connection.put(uri)

    def unblock(self, user_id):
        """Removes a block, if one exists, on the user with the given user ID. Unblocked users are
        allowed to log in and restore.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#UserBlock-delete>`__

        Args:
            user_id (int): An ID for a user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/UserBlock/{}".format(user_id)
        return self._connection.delete(uri)

    def deactivate(self, user_id, block_user=None):
        """Deactivates the user with the given user ID.
        Backups discontinue for a deactivated user, and their archives go to cold storage.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#UserDeactivation-put>`__

        Args:
            user_id (int): An ID for a user.
            block_user (bool, optional): Blocks the user upon deactivation. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/UserDeactivation/{}".format(user_id)
        data = {u"blockUser": block_user}
        return self._connection.put(uri, json=data)

    def reactivate(self, user_id, unblock_user=None):
        """Reactivates the user with the given ID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#UserDeactivation-delete>`__

        Args:
            user_id (int): An ID for a user.
            unblock_user (bool, optional): Whether or not to unblock the user. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/UserDeactivation/{}".format(user_id)
        params = {u"unblockUser": unblock_user}
        return self._connection.delete(uri, params=params)

    def change_org_assignment(self, user_id, org_id):
        """Assigns a user to a different organization.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#UserMoveProcess-post>`__

        Args:
            user_id (int): An ID for a user.
            org_id (int): An ID for the organization to move the user to.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/UserMoveProcess"
        data = {u"userId": user_id, u"parentOrgId": org_id}
        return self._connection.post(uri, json=data)

    def get_available_roles(self):
        """Report the list of roles that are available for the authenticated user to
        assign to other users.
        `V4 REST Documentation <https://console.us.code42.com/swagger/#/role/Role_View>`__

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/v4/role/view"
        return self._connection.get(uri)

    def get_roles(self, user_id):
        """Return the list of roles that are currently assigned to the given user.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#UserRole-get>`__

        Args:
            user_id (int): An ID for a user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/UserRole/{}".format(user_id)
        return self._connection.get(uri)

    def add_role(self, user_id, role_name):
        """Adds a role to a user.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#UserRole-post>`__

        Args:
            user_id (int): An ID for a user.
            role_name (str): The name of the role to assign to the user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/UserRole"
        data = {u"userId": user_id, u"roleName": role_name}
        return self._connection.post(uri, json=data)

    def remove_role(self, user_id, role_name):
        """Removes a role from a user.
        REST Documentation <https://console.us.code42.com/apidocviewer/#UserRole-delete>`__

        Args:
            user_id (int): An ID for a user.
            role_name (str): The name of the role to unassign from the user.

        Returns:
            :class:`py42.response.Py42Response`
        """

        # use quote instead of params here so that %20 is used instead of + for spaces.
        role_name = quote(role_name)
        uri = u"/api/UserRole?userId={}&roleName={}".format(user_id, role_name)
        return self._connection.delete(uri)



from py42.sdk.queries.fileevents.file_event_query import FileEventFilterStringField


class Printer(FileEventFilterStringField):
    """Class that filters events by printer name."""

    _term = u"printerName"


class PrintJobName(FileEventFilterStringField):
    """Class that filters events by print job name."""

    _term = u"printJobName"



class UserContext(object):
    """An object representing the currently logged in user."""

    def __init__(self, administration_client):
        self._administration_client = administration_client
        self._tenant_id = None

    def get_current_tenant_id(self):
        """Gets the currently signed in user's tenant ID."""
        if self._tenant_id is None:
            self._tenant_id = self._get_tenant_id()
        return self._tenant_id

    def _get_tenant_id(self):
        response = self._administration_client.get_current_tenant()
        return response[u"tenantUid"]



from py42.sdk.queries.fileevents.filters.activity_filter import *
from py42.sdk.queries.fileevents.filters.cloud_filter import *
from py42.sdk.queries.fileevents.filters.device_filter import *
from py42.sdk.queries.fileevents.filters.email_filter import *
from py42.sdk.queries.fileevents.filters.event_filter import *
from py42.sdk.queries.fileevents.filters.exposure_filter import *
from py42.sdk.queries.fileevents.filters.file_filter import *
from py42.sdk.queries.fileevents.filters.print_filter import *
